@@ -19,21 +19,17 @@ package csi
 import (
 	"context"
 	"fmt"
-	"io/ioutil"
 	"math/rand"
 	"os"
-	"path"
 	"path/filepath"
+	"reflect"
+	goruntime "runtime"
 	"testing"
 	"time"
 
-	"reflect"
-
 	"github.com/google/go-cmp/cmp"
 	"github.com/stretchr/testify/assert"
-
 	authenticationv1 "k8s.io/api/authentication/v1"
-	api "k8s.io/api/core/v1"
 	corev1 "k8s.io/api/core/v1"
 	storage "k8s.io/api/storage/v1"
 	meta "k8s.io/apimachinery/pkg/apis/meta/v1"
@@ -63,7 +59,7 @@ var (
 	testAccount = "test-service-account"
 )
 
-func prepareVolumeInfoFile(mountPath string, plug *csiPlugin, specVolumeName, volumeID, driverName, lifecycleMode string) error {
+func prepareVolumeInfoFile(mountPath string, plug *csiPlugin, specVolumeName, volumeID, driverName, lifecycleMode, seLinuxMountContext string) error {
 	nodeName := string(plug.host.GetNodeName())
 	volData := map[string]string{
 		volDataKey.specVolID:           specVolumeName,
@@ -72,6 +68,7 @@ func prepareVolumeInfoFile(mountPath string, plug *csiPlugin, specVolumeName, vo
 		volDataKey.nodeName:            nodeName,
 		volDataKey.attachmentID:        getAttachmentName(volumeID, driverName, nodeName),
 		volDataKey.volumeLifecycleMode: lifecycleMode,
+		volDataKey.seLinuxMountContext: seLinuxMountContext,
 	}
 	if err := os.MkdirAll(mountPath, 0755); err != nil {
 		return fmt.Errorf("failed to create dir for volume info file: %s", err)
@@ -110,7 +107,7 @@ func TestMounterGetPath(t *testing.T) {
 		spec := volume.NewSpecFromPersistentVolume(pv, pv.Spec.PersistentVolumeSource.CSI.ReadOnly)
 		mounter, err := plug.NewMounter(
 			spec,
-			&api.Pod{ObjectMeta: meta.ObjectMeta{UID: testPodUID, Namespace: testns}},
+			&corev1.Pod{ObjectMeta: meta.ObjectMeta{UID: testPodUID, Namespace: testns}},
 			volume.VolumeOptions{},
 		)
 		if err != nil {
@@ -128,11 +125,13 @@ func TestMounterGetPath(t *testing.T) {
 
 func TestMounterSetUp(t *testing.T) {
 	tests := []struct {
-		name                  string
-		driver                string
-		volumeContext         map[string]string
-		expectedVolumeContext map[string]string
-		csiInlineVolume       bool
+		name                     string
+		driver                   string
+		volumeContext            map[string]string
+		seLinuxLabel             string
+		enableSELinuxFeatureGate bool
+		expectedSELinuxContext   string
+		expectedVolumeContext    map[string]string
 	}{
 		{
 			name:                  "no pod info",
@@ -162,20 +161,53 @@ func TestMounterSetUp(t *testing.T) {
 			name:                  "add pod info",
 			driver:                "info",
 			volumeContext:         nil,
-			expectedVolumeContext: map[string]string{"csi.storage.k8s.io/pod.uid": "test-pod", "csi.storage.k8s.io/serviceAccount.name": "test-service-account", "csi.storage.k8s.io/pod.name": "test-pod", "csi.storage.k8s.io/pod.namespace": "test-ns"},
+			expectedVolumeContext: map[string]string{"csi.storage.k8s.io/pod.uid": "test-pod", "csi.storage.k8s.io/serviceAccount.name": "test-service-account", "csi.storage.k8s.io/pod.name": "test-pod", "csi.storage.k8s.io/pod.namespace": "test-ns", "csi.storage.k8s.io/ephemeral": "false"},
 		},
 		{
 			name:                  "add pod info -> keep existing volumeContext",
 			driver:                "info",
 			volumeContext:         map[string]string{"foo": "bar"},
-			expectedVolumeContext: map[string]string{"foo": "bar", "csi.storage.k8s.io/pod.uid": "test-pod", "csi.storage.k8s.io/serviceAccount.name": "test-service-account", "csi.storage.k8s.io/pod.name": "test-pod", "csi.storage.k8s.io/pod.namespace": "test-ns"},
+			expectedVolumeContext: map[string]string{"foo": "bar", "csi.storage.k8s.io/pod.uid": "test-pod", "csi.storage.k8s.io/serviceAccount.name": "test-service-account", "csi.storage.k8s.io/pod.name": "test-pod", "csi.storage.k8s.io/pod.namespace": "test-ns", "csi.storage.k8s.io/ephemeral": "false"},
 		},
 		{
 			name:                  "CSIInlineVolume pod info",
 			driver:                "info",
 			volumeContext:         nil,
 			expectedVolumeContext: map[string]string{"csi.storage.k8s.io/pod.uid": "test-pod", "csi.storage.k8s.io/serviceAccount.name": "test-service-account", "csi.storage.k8s.io/pod.name": "test-pod", "csi.storage.k8s.io/pod.namespace": "test-ns", "csi.storage.k8s.io/ephemeral": "false"},
-			csiInlineVolume:       true,
+		},
+		{
+			name:                     "should include SELinux mount options, if feature-gate is enabled and driver supports it",
+			driver:                   "supports_selinux",
+			volumeContext:            nil,
+			seLinuxLabel:             "s0,c0",
+			expectedSELinuxContext:   "context=\"s0,c0\"",
+			enableSELinuxFeatureGate: true,
+			expectedVolumeContext:    nil,
+		},
+		{
+			name:                     "should not include selinux mount options, if feature gate is enabled but driver does not support it",
+			driver:                   "no_selinux",
+			seLinuxLabel:             "s0,c0",
+			volumeContext:            nil,
+			enableSELinuxFeatureGate: true,
+			expectedVolumeContext:    nil,
+		},
+		{
+			name:                     "should not include selinux mount option, if feature gate is enabled but CSIDriver does not exist",
+			driver:                   "not_found_selinux",
+			seLinuxLabel:             "s0,c0",
+			volumeContext:            nil,
+			enableSELinuxFeatureGate: true,
+			expectedVolumeContext:    nil,
+		},
+		{
+			name:                     "should not include selinux mount options, if feature gate is enabled, driver supports it, but Pod does not have it",
+			driver:                   "supports_selinux",
+			seLinuxLabel:             "",
+			expectedSELinuxContext:   "", // especially make sure the volume plugin does not use -o context="", that is an invalid value
+			volumeContext:            nil,
+			enableSELinuxFeatureGate: true,
+			expectedVolumeContext:    nil,
 		},
 	}
 
@@ -183,16 +215,17 @@ func TestMounterSetUp(t *testing.T) {
 	currentPodInfoMount := true
 	for _, test := range tests {
 		t.Run(test.name, func(t *testing.T) {
-			// Modes must be set if (and only if) CSIInlineVolume is enabled.
-			var modes []storage.VolumeLifecycleMode
-			defer featuregatetesting.SetFeatureGateDuringTest(t, utilfeature.DefaultFeatureGate, features.CSIInlineVolume, test.csiInlineVolume)()
-			if test.csiInlineVolume {
-				modes = append(modes, storage.VolumeLifecyclePersistent)
+			defer featuregatetesting.SetFeatureGateDuringTest(t, utilfeature.DefaultFeatureGate, features.SELinuxMountReadWriteOncePod, test.enableSELinuxFeatureGate)()
+
+			modes := []storage.VolumeLifecycleMode{
+				storage.VolumeLifecyclePersistent,
 			}
 			fakeClient := fakeclient.NewSimpleClientset(
 				getTestCSIDriver("no-info", &noPodMountInfo, nil, modes),
 				getTestCSIDriver("info", &currentPodInfoMount, nil, modes),
 				getTestCSIDriver("nil", nil, nil, modes),
+				getTestCSIDriver("supports_selinux", &noPodMountInfo, nil, modes),
+				getTestCSIDriver("no_selinux", &noPodMountInfo, nil, modes),
 			)
 			plug, tmpDir := newTestPlugin(t, fakeClient)
 			defer os.RemoveAll(tmpDir)
@@ -205,9 +238,9 @@ func TestMounterSetUp(t *testing.T) {
 
 			mounter, err := plug.NewMounter(
 				volume.NewSpecFromPersistentVolume(pv, pv.Spec.PersistentVolumeSource.CSI.ReadOnly),
-				&api.Pod{
+				&corev1.Pod{
 					ObjectMeta: meta.ObjectMeta{UID: testPodUID, Namespace: testns, Name: testPod},
-					Spec: api.PodSpec{
+					Spec: corev1.PodSpec{
 						ServiceAccountName: testAccount,
 					},
 				},
@@ -252,10 +285,20 @@ func TestMounterSetUp(t *testing.T) {
 			var mounterArgs volume.MounterArgs
 			fsGroup := int64(2000)
 			mounterArgs.FsGroup = &fsGroup
+
+			if test.seLinuxLabel != "" {
+				mounterArgs.SELinuxLabel = test.seLinuxLabel
+			}
+
+			expectedMountOptions := pv.Spec.MountOptions
+
+			if test.expectedSELinuxContext != "" {
+				expectedMountOptions = append(expectedMountOptions, test.expectedSELinuxContext)
+			}
+
 			if err := csiMounter.SetUp(mounterArgs); err != nil {
 				t.Fatalf("mounter.Setup failed: %v", err)
 			}
-
 			//Test the default value of file system type is not overridden
 			if len(csiMounter.spec.PersistentVolume.Spec.CSI.FSType) != 0 {
 				t.Errorf("default value of file system type was overridden by type %s", csiMounter.spec.PersistentVolume.Spec.CSI.FSType)
@@ -279,8 +322,8 @@ func TestMounterSetUp(t *testing.T) {
 			if vol.Path != csiMounter.GetPath() {
 				t.Errorf("csi server expected path %s, got %s", csiMounter.GetPath(), vol.Path)
 			}
-			if !reflect.DeepEqual(vol.MountFlags, pv.Spec.MountOptions) {
-				t.Errorf("csi server expected mount options %v, got %v", pv.Spec.MountOptions, vol.MountFlags)
+			if !reflect.DeepEqual(vol.MountFlags, expectedMountOptions) {
+				t.Errorf("csi server expected mount options %v, got %v", expectedMountOptions, vol.MountFlags)
 			}
 			if !reflect.DeepEqual(vol.VolumeContext, test.expectedVolumeContext) {
 				t.Errorf("csi server expected volumeContext %+v, got %+v", test.expectedVolumeContext, vol.VolumeContext)
@@ -326,20 +369,21 @@ func TestMounterSetUpSimple(t *testing.T) {
 	defer os.RemoveAll(tmpDir)
 
 	testCases := []struct {
-		name       string
-		podUID     types.UID
-		mode       storage.VolumeLifecycleMode
-		fsType     string
-		options    []string
-		spec       func(string, []string) *volume.Spec
-		shouldFail bool
+		name                 string
+		podUID               types.UID
+		mode                 storage.VolumeLifecycleMode
+		fsType               string
+		options              []string
+		spec                 func(string, []string) *volume.Spec
+		newMounterShouldFail bool
+		setupShouldFail      bool
 	}{
 		{
-			name:       "setup with ephemeral source",
-			podUID:     types.UID(fmt.Sprintf("%08X", rand.Uint64())),
-			mode:       storage.VolumeLifecycleEphemeral,
-			fsType:     "ext4",
-			shouldFail: true,
+			name:            "setup with ephemeral source",
+			podUID:          types.UID(fmt.Sprintf("%08X", rand.Uint64())),
+			mode:            storage.VolumeLifecycleEphemeral,
+			fsType:          "ext4",
+			setupShouldFail: true,
 			spec: func(fsType string, options []string) *volume.Spec {
 				volSrc := makeTestVol("pv1", testDriver)
 				volSrc.CSI.FSType = &fsType
@@ -367,9 +411,9 @@ func TestMounterSetUpSimple(t *testing.T) {
 			},
 		},
 		{
-			name:       "setup with missing spec",
-			shouldFail: true,
-			spec:       func(fsType string, options []string) *volume.Spec { return nil },
+			name:                 "setup with missing spec",
+			newMounterShouldFail: true,
+			spec:                 func(fsType string, options []string) *volume.Spec { return nil },
 		},
 	}
 
@@ -378,14 +422,14 @@ func TestMounterSetUpSimple(t *testing.T) {
 		t.Run(tc.name, func(t *testing.T) {
 			mounter, err := plug.NewMounter(
 				tc.spec(tc.fsType, tc.options),
-				&api.Pod{ObjectMeta: meta.ObjectMeta{UID: tc.podUID, Namespace: testns}},
+				&corev1.Pod{ObjectMeta: meta.ObjectMeta{UID: tc.podUID, Namespace: testns}},
 				volume.VolumeOptions{},
 			)
-			if tc.shouldFail && err != nil {
+			if tc.newMounterShouldFail && err != nil {
 				t.Log(err)
 				return
 			}
-			if !tc.shouldFail && err != nil {
+			if !tc.newMounterShouldFail && err != nil {
 				t.Fatal("unexpected error:", err)
 			}
 			if mounter == nil {
@@ -395,7 +439,7 @@ func TestMounterSetUpSimple(t *testing.T) {
 			csiMounter := mounter.(*csiMountMgr)
 			csiMounter.csiClient = setupClient(t, true)
 
-			if csiMounter.volumeLifecycleMode != storage.VolumeLifecyclePersistent {
+			if csiMounter.volumeLifecycleMode != tc.mode {
 				t.Fatal("unexpected volume mode: ", csiMounter.volumeLifecycleMode)
 			}
 
@@ -407,8 +451,13 @@ func TestMounterSetUpSimple(t *testing.T) {
 			}
 
 			// Mounter.SetUp()
-			if err := csiMounter.SetUp(volume.MounterArgs{}); err != nil {
-				t.Fatalf("mounter.Setup failed: %v", err)
+			err = csiMounter.SetUp(volume.MounterArgs{})
+			if tc.setupShouldFail && err != nil {
+				t.Log(err)
+				return
+			}
+			if !tc.setupShouldFail && err != nil {
+				t.Fatal("unexpected error:", err)
 			}
 
 			// ensure call went all the way
@@ -525,7 +574,7 @@ func TestMounterSetupWithStatusTracking(t *testing.T) {
 			podUID: types.UID(fmt.Sprintf("%08X", rand.Uint64())),
 			spec: func(fsType string, options []string) *volume.Spec {
 				pv := makeTestPV("pv5", 20, testDriver, "vol6")
-				pv.Spec.PersistentVolumeSource.CSI.NodePublishSecretRef = &api.SecretReference{
+				pv.Spec.PersistentVolumeSource.CSI.NodePublishSecretRef = &corev1.SecretReference{
 					Name:      "foo",
 					Namespace: "default",
 				}
@@ -542,7 +591,7 @@ func TestMounterSetupWithStatusTracking(t *testing.T) {
 		t.Run(tc.name, func(t *testing.T) {
 			mounter, err := plug.NewMounter(
 				tc.spec("ext4", []string{}),
-				&api.Pod{ObjectMeta: meta.ObjectMeta{UID: tc.podUID, Namespace: testns}},
+				&corev1.Pod{ObjectMeta: meta.ObjectMeta{UID: tc.podUID, Namespace: testns}},
 				volume.VolumeOptions{},
 			)
 			if err != nil {
@@ -582,8 +631,6 @@ func TestMounterSetupWithStatusTracking(t *testing.T) {
 }
 
 func TestMounterSetUpWithInline(t *testing.T) {
-	defer featuregatetesting.SetFeatureGateDuringTest(t, utilfeature.DefaultFeatureGate, features.CSIInlineVolume, true)()
-
 	testCases := []struct {
 		name       string
 		podUID     types.UID
@@ -645,7 +692,7 @@ func TestMounterSetUpWithInline(t *testing.T) {
 		t.Run(tc.name, func(t *testing.T) {
 			mounter, err := plug.NewMounter(
 				tc.spec(tc.fsType, tc.options),
-				&api.Pod{ObjectMeta: meta.ObjectMeta{UID: tc.podUID, Namespace: testns}},
+				&corev1.Pod{ObjectMeta: meta.ObjectMeta{UID: tc.podUID, Namespace: testns}},
 				volume.VolumeOptions{},
 			)
 			if tc.shouldFail && err != nil {
@@ -730,29 +777,28 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 
 	testCases := []struct {
 		name                           string
-		accessModes                    []api.PersistentVolumeAccessMode
+		accessModes                    []corev1.PersistentVolumeAccessMode
 		readOnly                       bool
 		fsType                         string
 		setFsGroup                     bool
 		fsGroup                        int64
 		driverFSGroupPolicy            bool
 		supportMode                    storage.FSGroupPolicy
-		delegateFSGroupFeatureGate     bool
 		driverSupportsVolumeMountGroup bool
 		expectedFSGroupInNodePublish   string
 	}{
 		{
 			name: "default fstype, with no fsgroup (should not apply fsgroup)",
-			accessModes: []api.PersistentVolumeAccessMode{
-				api.ReadWriteOnce,
+			accessModes: []corev1.PersistentVolumeAccessMode{
+				corev1.ReadWriteOnce,
 			},
 			readOnly: false,
 			fsType:   "",
 		},
 		{
 			name: "default fstype  with fsgroup (should not apply fsgroup)",
-			accessModes: []api.PersistentVolumeAccessMode{
-				api.ReadWriteOnce,
+			accessModes: []corev1.PersistentVolumeAccessMode{
+				corev1.ReadWriteOnce,
 			},
 			readOnly:   false,
 			fsType:     "",
@@ -761,9 +807,9 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 		},
 		{
 			name: "fstype, fsgroup, RWM, ROM provided (should not apply fsgroup)",
-			accessModes: []api.PersistentVolumeAccessMode{
-				api.ReadWriteMany,
-				api.ReadOnlyMany,
+			accessModes: []corev1.PersistentVolumeAccessMode{
+				corev1.ReadWriteMany,
+				corev1.ReadOnlyMany,
 			},
 			fsType:     "ext4",
 			setFsGroup: true,
@@ -771,8 +817,8 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 		},
 		{
 			name: "fstype, fsgroup, RWO, but readOnly (should not apply fsgroup)",
-			accessModes: []api.PersistentVolumeAccessMode{
-				api.ReadWriteOnce,
+			accessModes: []corev1.PersistentVolumeAccessMode{
+				corev1.ReadWriteOnce,
 			},
 			readOnly:   true,
 			fsType:     "ext4",
@@ -781,8 +827,8 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 		},
 		{
 			name: "fstype, fsgroup, RWO provided (should apply fsgroup)",
-			accessModes: []api.PersistentVolumeAccessMode{
-				api.ReadWriteOnce,
+			accessModes: []corev1.PersistentVolumeAccessMode{
+				corev1.ReadWriteOnce,
 			},
 			fsType:     "ext4",
 			setFsGroup: true,
@@ -790,8 +836,8 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 		},
 		{
 			name: "fstype, fsgroup, RWO provided, FSGroupPolicy ReadWriteOnceWithFSType (should apply fsgroup)",
-			accessModes: []api.PersistentVolumeAccessMode{
-				api.ReadWriteOnce,
+			accessModes: []corev1.PersistentVolumeAccessMode{
+				corev1.ReadWriteOnce,
 			},
 			fsType:              "ext4",
 			setFsGroup:          true,
@@ -801,8 +847,8 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 		},
 		{
 			name: "default fstype with no fsgroup, FSGroupPolicy ReadWriteOnceWithFSType (should not apply fsgroup)",
-			accessModes: []api.PersistentVolumeAccessMode{
-				api.ReadWriteOnce,
+			accessModes: []corev1.PersistentVolumeAccessMode{
+				corev1.ReadWriteOnce,
 			},
 			readOnly:            false,
 			fsType:              "",
@@ -811,8 +857,8 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 		},
 		{
 			name: "default fstype with fsgroup, FSGroupPolicy ReadWriteOnceWithFSType (should not apply fsgroup)",
-			accessModes: []api.PersistentVolumeAccessMode{
-				api.ReadWriteOnce,
+			accessModes: []corev1.PersistentVolumeAccessMode{
+				corev1.ReadWriteOnce,
 			},
 			readOnly:            false,
 			fsType:              "",
@@ -823,8 +869,8 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 		},
 		{
 			name: "fstype, fsgroup, RWO provided, readonly, FSGroupPolicy ReadWriteOnceWithFSType (should not apply fsgroup)",
-			accessModes: []api.PersistentVolumeAccessMode{
-				api.ReadWriteOnce,
+			accessModes: []corev1.PersistentVolumeAccessMode{
+				corev1.ReadWriteOnce,
 			},
 			readOnly:            true,
 			fsType:              "ext4",
@@ -835,8 +881,8 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 		},
 		{
 			name: "fstype, fsgroup, RWX provided, FSGroupPolicy ReadWriteOnceWithFSType (should not apply fsgroup)",
-			accessModes: []api.PersistentVolumeAccessMode{
-				api.ReadWriteMany,
+			accessModes: []corev1.PersistentVolumeAccessMode{
+				corev1.ReadWriteMany,
 			},
 			readOnly:            false,
 			fsType:              "ext4",
@@ -847,8 +893,8 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 		},
 		{
 			name: "fstype, fsgroup, RWO provided, FSGroupPolicy None (should not apply fsgroup)",
-			accessModes: []api.PersistentVolumeAccessMode{
-				api.ReadWriteOnce,
+			accessModes: []corev1.PersistentVolumeAccessMode{
+				corev1.ReadWriteOnce,
 			},
 			fsType:              "ext4",
 			setFsGroup:          true,
@@ -858,8 +904,8 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 		},
 		{
 			name: "fstype, fsgroup, RWO provided, readOnly, FSGroupPolicy File (should apply fsgroup)",
-			accessModes: []api.PersistentVolumeAccessMode{
-				api.ReadWriteOnce,
+			accessModes: []corev1.PersistentVolumeAccessMode{
+				corev1.ReadWriteOnce,
 			},
 			readOnly:            true,
 			fsType:              "ext4",
@@ -869,47 +915,33 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 			supportMode:         storage.FileFSGroupPolicy,
 		},
 		{
-			name:                           "fsgroup provided, DelegateFSGroupToCSIDriver feature enabled, driver supports volume mount group; expect fsgroup to be passed to NodePublishVolume",
+			name:                           "fsgroup provided, driver supports volume mount group; expect fsgroup to be passed to NodePublishVolume",
 			fsType:                         "ext4",
 			setFsGroup:                     true,
 			fsGroup:                        3000,
-			delegateFSGroupFeatureGate:     true,
 			driverSupportsVolumeMountGroup: true,
 			expectedFSGroupInNodePublish:   "3000",
 		},
 		{
-			name:                           "fsgroup not provided, DelegateFSGroupToCSIDriver feature enabled, driver supports volume mount group; expect fsgroup not to be passed to NodePublishVolume",
+			name:                           "fsgroup not provided, driver supports volume mount group; expect fsgroup not to be passed to NodePublishVolume",
 			fsType:                         "ext4",
 			setFsGroup:                     false,
-			delegateFSGroupFeatureGate:     true,
 			driverSupportsVolumeMountGroup: true,
 			expectedFSGroupInNodePublish:   "",
 		},
 		{
-			name:                           "fsgroup provided, DelegateFSGroupToCSIDriver feature enabled, driver does not support volume mount group; expect fsgroup not to be passed to NodePublishVolume",
+			name:                           "fsgroup provided, driver does not support volume mount group; expect fsgroup not to be passed to NodePublishVolume",
 			fsType:                         "ext4",
 			setFsGroup:                     true,
 			fsGroup:                        3000,
-			delegateFSGroupFeatureGate:     true,
 			driverSupportsVolumeMountGroup: false,
 			expectedFSGroupInNodePublish:   "",
 		},
-		{
-			name:                           "fsgroup provided, DelegateFSGroupToCSIDriver feature disabled, driver supports volume mount group; expect fsgroup not to be passed to NodePublishVolume",
-			fsType:                         "ext4",
-			setFsGroup:                     true,
-			fsGroup:                        3000,
-			delegateFSGroupFeatureGate:     false,
-			driverSupportsVolumeMountGroup: true,
-			expectedFSGroupInNodePublish:   "",
-		},
 	}
 
 	for i, tc := range testCases {
 		t.Logf("Running test %s", tc.name)
 
-		defer featuregatetesting.SetFeatureGateDuringTest(t, utilfeature.DefaultFeatureGate, features.DelegateFSGroupToCSIDriver, tc.delegateFSGroupFeatureGate)()
-
 		volName := fmt.Sprintf("test-vol-%d", i)
 		registerFakePlugin(testDriver, "endpoint", []string{"1.0.0"}, t)
 		pv := makeTestPV("test-pv", 10, testDriver, volName)
@@ -924,7 +956,7 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 
 		mounter, err := plug.NewMounter(
 			spec,
-			&api.Pod{ObjectMeta: meta.ObjectMeta{UID: testPodUID, Namespace: testns}},
+			&corev1.Pod{ObjectMeta: meta.ObjectMeta{UID: testPodUID, Namespace: testns}},
 			volume.VolumeOptions{},
 		)
 		if err != nil {
@@ -936,9 +968,6 @@ func TestMounterSetUpWithFSGroup(t *testing.T) {
 		}
 
 		csiMounter := mounter.(*csiMountMgr)
-		if tc.driverFSGroupPolicy {
-			csiMounter.fsGroupPolicy = tc.supportMode
-		}
 		csiMounter.csiClient = setupClientWithVolumeMountGroup(t, true /* stageUnstageSet */, tc.driverSupportsVolumeMountGroup)
 
 		attachID := getAttachmentName(csiMounter.volumeID, string(csiMounter.driverName), string(plug.host.GetNodeName()))
@@ -985,19 +1014,25 @@ func TestUnmounterTeardown(t *testing.T) {
 	pv := makeTestPV("test-pv", 10, testDriver, testVol)
 
 	// save the data file prior to unmount
-	dir := filepath.Join(getTargetPath(testPodUID, pv.ObjectMeta.Name, plug.host), "/mount")
+	targetDir := getTargetPath(testPodUID, pv.ObjectMeta.Name, plug.host)
+	dir := filepath.Join(targetDir, "mount")
 	if err := os.MkdirAll(dir, 0755); err != nil && !os.IsNotExist(err) {
 		t.Errorf("failed to create dir [%s]: %v", dir, err)
 	}
 
 	// do a fake local mount
 	diskMounter := util.NewSafeFormatAndMountFromHost(plug.GetPluginName(), plug.host)
-	if err := diskMounter.FormatAndMount("/fake/device", dir, "testfs", nil); err != nil {
+	device := "/fake/device"
+	if goruntime.GOOS == "windows" {
+		// We need disk numbers on Windows.
+		device = "1"
+	}
+	if err := diskMounter.FormatAndMount(device, dir, "testfs", nil); err != nil {
 		t.Errorf("failed to mount dir [%s]: %v", dir, err)
 	}
 
 	if err := saveVolumeData(
-		path.Dir(dir),
+		targetDir,
 		volDataFileName,
 		map[string]string{
 			volDataKey.specVolID:  pv.ObjectMeta.Name,
@@ -1029,7 +1064,7 @@ func TestUnmounterTeardown(t *testing.T) {
 }
 
 func TestIsCorruptedDir(t *testing.T) {
-	existingMountPath, err := ioutil.TempDir(os.TempDir(), "blobfuse-csi-mount-test")
+	existingMountPath, err := os.MkdirTemp(os.TempDir(), "blobfuse-csi-mount-test")
 	if err != nil {
 		t.Fatalf("failed to create tmp dir: %v", err)
 	}
@@ -1059,7 +1094,6 @@ func TestIsCorruptedDir(t *testing.T) {
 }
 
 func TestPodServiceAccountTokenAttrs(t *testing.T) {
-	defer featuregatetesting.SetFeatureGateDuringTest(t, utilfeature.DefaultFeatureGate, features.CSIInlineVolume, true)()
 	scheme := runtime.NewScheme()
 	utilruntime.Must(pkgauthenticationv1.RegisterDefaults(scheme))
 	utilruntime.Must(pkgstoragev1.RegisterDefaults(scheme))
@@ -1143,9 +1177,9 @@ func TestPodServiceAccountTokenAttrs(t *testing.T) {
 			defer os.RemoveAll(tmpDir)
 			mounter, err := plug.NewMounter(
 				volume.NewSpecFromVolume(makeTestVol("test", testDriver)),
-				&api.Pod{
+				&corev1.Pod{
 					ObjectMeta: meta.ObjectMeta{UID: testPodUID, Namespace: testns, Name: testPod},
-					Spec: api.PodSpec{
+					Spec: corev1.PodSpec{
 						ServiceAccountName: testAccount,
 					},
 				},
@@ -1181,13 +1215,12 @@ func Test_csiMountMgr_supportsFSGroup(t *testing.T) {
 		plugin              *csiPlugin
 		driverName          csiDriverName
 		volumeLifecycleMode storage.VolumeLifecycleMode
-		fsGroupPolicy       storage.FSGroupPolicy
 		volumeID            string
 		specVolumeID        string
 		readOnly            bool
 		supportsSELinux     bool
 		spec                *volume.Spec
-		pod                 *api.Pod
+		pod                 *corev1.Pod
 		podUID              types.UID
 		publishContext      map[string]string
 		kubeVolHost         volume.KubeletVolumeHost
@@ -1257,9 +1290,9 @@ func Test_csiMountMgr_supportsFSGroup(t *testing.T) {
 				driverPolicy: storage.ReadWriteOnceWithFSTypeFSGroupPolicy,
 			},
 			fields: fields{
-				spec: volume.NewSpecFromPersistentVolume(&api.PersistentVolume{
-					Spec: api.PersistentVolumeSpec{
-						AccessModes: []api.PersistentVolumeAccessMode{},
+				spec: volume.NewSpecFromPersistentVolume(&corev1.PersistentVolume{
+					Spec: corev1.PersistentVolumeSpec{
+						AccessModes: []corev1.PersistentVolumeAccessMode{},
 					},
 				}, true),
 			},
@@ -1273,9 +1306,9 @@ func Test_csiMountMgr_supportsFSGroup(t *testing.T) {
 				driverPolicy: storage.ReadWriteOnceWithFSTypeFSGroupPolicy,
 			},
 			fields: fields{
-				spec: volume.NewSpecFromPersistentVolume(&api.PersistentVolume{
-					Spec: api.PersistentVolumeSpec{
-						AccessModes: []api.PersistentVolumeAccessMode{api.ReadWriteOnce},
+				spec: volume.NewSpecFromPersistentVolume(&corev1.PersistentVolume{
+					Spec: corev1.PersistentVolumeSpec{
+						AccessModes: []corev1.PersistentVolumeAccessMode{corev1.ReadWriteOnce},
 					},
 				}, true),
 			},
@@ -1289,9 +1322,9 @@ func Test_csiMountMgr_supportsFSGroup(t *testing.T) {
 				driverPolicy: storage.ReadWriteOnceWithFSTypeFSGroupPolicy,
 			},
 			fields: fields{
-				spec: volume.NewSpecFromVolume(&api.Volume{
-					VolumeSource: api.VolumeSource{
-						CSI: &api.CSIVolumeSource{
+				spec: volume.NewSpecFromVolume(&corev1.Volume{
+					VolumeSource: corev1.VolumeSource{
+						CSI: &corev1.CSIVolumeSource{
 							Driver: testDriver,
 						},
 					},
@@ -1307,11 +1340,10 @@ func Test_csiMountMgr_supportsFSGroup(t *testing.T) {
 				plugin:              tt.fields.plugin,
 				driverName:          tt.fields.driverName,
 				volumeLifecycleMode: tt.fields.volumeLifecycleMode,
-				fsGroupPolicy:       tt.fields.fsGroupPolicy,
 				volumeID:            tt.fields.volumeID,
 				specVolumeID:        tt.fields.specVolumeID,
 				readOnly:            tt.fields.readOnly,
-				supportsSELinux:     tt.fields.supportsSELinux,
+				needSELinuxRelabel:  tt.fields.supportsSELinux,
 				spec:                tt.fields.spec,
 				pod:                 tt.fields.pod,
 				podUID:              tt.fields.podUID,
@@ -1325,3 +1357,64 @@ func Test_csiMountMgr_supportsFSGroup(t *testing.T) {
 		})
 	}
 }
+
+func TestMounterGetFSGroupPolicy(t *testing.T) {
+	defaultPolicy := storage.ReadWriteOnceWithFSTypeFSGroupPolicy
+	testCases := []struct {
+		name                  string
+		defined               bool
+		expectedFSGroupPolicy storage.FSGroupPolicy
+	}{
+		{
+			name:                  "no FSGroupPolicy defined, expect default",
+			defined:               false,
+			expectedFSGroupPolicy: storage.ReadWriteOnceWithFSTypeFSGroupPolicy,
+		},
+		{
+			name:                  "File FSGroupPolicy defined, expect File",
+			defined:               true,
+			expectedFSGroupPolicy: storage.FileFSGroupPolicy,
+		},
+		{
+			name:                  "None FSGroupPolicy defined, expected None",
+			defined:               true,
+			expectedFSGroupPolicy: storage.NoneFSGroupPolicy,
+		},
+	}
+	for _, tc := range testCases {
+		t.Logf("testing: %s", tc.name)
+		// Define the driver and set the FSGroupPolicy
+		driver := getTestCSIDriver(testDriver, nil, nil, nil)
+		if tc.defined {
+			driver.Spec.FSGroupPolicy = &tc.expectedFSGroupPolicy
+		} else {
+			driver.Spec.FSGroupPolicy = &defaultPolicy
+		}
+
+		// Create the client and register the resources
+		fakeClient := fakeclient.NewSimpleClientset(driver)
+		plug, tmpDir := newTestPlugin(t, fakeClient)
+		defer os.RemoveAll(tmpDir)
+		registerFakePlugin(testDriver, "endpoint", []string{"1.3.0"}, t)
+
+		mounter, err := plug.NewMounter(
+			volume.NewSpecFromPersistentVolume(makeTestPV("test.vol.id", 20, testDriver, "testvol-handle1"), true),
+			&corev1.Pod{ObjectMeta: meta.ObjectMeta{UID: "1", Namespace: testns}},
+			volume.VolumeOptions{},
+		)
+		if err != nil {
+			t.Fatalf("Error creating a new mounter: %s", err)
+		}
+
+		csiMounter := mounter.(*csiMountMgr)
+
+		// Check to see if we can obtain the CSIDriver, along with examining its FSGroupPolicy
+		fsGroup, err := csiMounter.getFSGroupPolicy()
+		if err != nil {
+			t.Fatalf("Error attempting to obtain FSGroupPolicy: %v", err)
+		}
+		if fsGroup != *driver.Spec.FSGroupPolicy {
+			t.Fatalf("FSGroupPolicy doesn't match expected value: %v, %v", fsGroup, tc.expectedFSGroupPolicy)
+		}
+	}
+}
